#  ncgmp09_TranslateToShape.py
#
#  Converts an NCGMP09-style ArcGIS geodatabase to 
#    open file format
#    	 shape files, .dbf files, and pipe-delimited text files,
#    	 without loss of information.  Field renaming is documented in 
#    	 output file logfile.txt
#    simple shapefile format
#      basic map information in flat shapefiles, with much repetition 
#      of attribute information, long fields truncated, and much 
#      information lost. Field renaming is documented in output 
#	 file logfile.txt
#
#  Ralph Haugerud, USGS, Seattle
#    rhaugerud@usgs.gov

import arcgisscripting, sys, os, glob, time

versionString = 'version of 8 October 2009'
Is93 = True
#  Code below has a couple of places that are dependent 
#  ArcGIS 9.3 or higher. These could be rewritten to work
#  with earlier versions

debug = False

# equivalentFraction is used to rank ProportionTerms from most 
#  abundant to least
equivalentFraction =   {'all':1.0,
			'only part':1.0,
			'dominant':0.6,
			'major':0.5,
			'significant':0.4,
			'subordinate':0.3,
			'minor':0.25,
			'trace':0.05,
			'rare':0.02,
			'variable':0.01,
			'present':0.0}

def usage():
	print """USAGE: ncgmp09_TranslateToShp.py  <geodatabase> <outputWorkspace>
  where <geodatabase> must be an existing ArcGIS geodatabase.
  <geodatabase> may be a personal or file geodatabase, and the 
  .gdb or .mdb extension must be included.
  Output is written to directories <geodatabase (no extension)>-simple
  and <geodatabase (no extension)>-open in <outputWorkspace>. Output 
  directories, if they already exist, will be overwritten.
"""

def addMsgAndPrint(gp,msg, severity=0): 
    # prints msg to screen and adds msg to the geoprocessor (in case this is run as a tool) 
    # print msg 
    try: 
        for string in msg.split('\n'): 
            # Add appropriate geoprocessing message 
            if severity == 0: 
                gp.AddMessage(string) 
            elif severity == 1: 
                gp.AddWarning(string) 
            elif severity == 2: 
                gp.AddError(string) 
    except: 
        pass 

def remapFieldName(name):
    name2 = name.replace('And','')
    name2 = name2.replace('Of','')
    name2 = name2.replace('Unit','Un')
    name2 = name2.replace('Source','Src')
    name2 = name2.replace('Shape','Shp')
    name2 = name2.replace('Hierarchy','H')
    name2 = name2.replace('Description','Descript')
    name2 = name2.replace('AreaFill','')
    newName = ''
    for i in range(0,len(name2)):
        if name2[i] == name2[i].upper():
            newName = newName + name2[i]
            j = 1
        else:
            j = j+1
            if j < 4:
                newName = newName + name2[i]
    if len(newName) > 10:
        if newName[1:3] == newName[1:3].lower():
            newName = newName[0]+newName[3:]
    if len(newName) > 10:
        if newName[3:5] == newName[3:5].lower():
            newName = newName[0:2]+newName[5:]
    if len(newName) > 10:
        print name, newName
    return newName	

def dumpTable(gp,fc,outName,isSpatial,outputDir,logfile,isOpen):
    dumpString = '  Dumping '+outName+'...'
    if isSpatial: dumpString = '  '+dumpString
    addMsgAndPrint(gp,dumpString)
    if isSpatial:
        logfile.write('  feature class '+fc+' dumped to shapefile '+outName+'\n')
    else:
        logfile.write('  table '+fc+' dumped to table '+outName+'\n')
    logfile.write('    field name remapping: \n')
    # describe
    fields = gp.ListFields(fc)
    longFields = []
    shortFieldName = {}
    for field in fields:
        # translate field names
        #  NEED TO FIX TO DEAL WITH DescriptionOfMapUnits_ and DataSources_
        fName = field.Name
        for prefix in ('DescriptionOfMapUnits','DataSources'):
            if fc <> prefix and fName.find(prefix) == 0:
                fName = fName[len(prefix)+1:]
        if len(fName) > 10:
            shortFieldName[field.Name] = remapFieldName(fName)
            # write field name translation to logfile
            logfile.write('      '+field.Name+' > '+shortFieldName[field.Name]+'\n')
        else:
            shortFieldName[field.Name] = fName
        if field.Length > 255:
            longFields.append(field.Name)
    # setup field mapping
    fms = gp.CreateObject("FieldMappings")
    # and add the table to generate a field map object for each field
    fms.AddTable(fc)
    i = 0
    while i < fms.fieldcount:
        # one fieldmap object per field
        fm = fms.getfieldmap(i)
        # get the output field object, really a copy of it
        fldout = fm.outputfield
        # set the name for the output field through the lookup dictionary
        fldout.name = shortFieldName[fm.getinputfieldname(0)]
        # re-apply the field to the fieldmap object
        fm.outputfield = fldout
        # and replace the existing field map at index i with the updated one.
        fms.replacefieldmap(i, fm)
        i = i + 1
    # export to shapefile (named prefix+name)
    if isSpatial:
        if debug:  print 'dumping ',fc,outputDir,outName
        gp.FeatureClassToFeatureClass_conversion(fc,outputDir,outName,'#',fms)
    else:
        gp.TableToTable_conversion(fc,outputDir,outName,'#',fms)
    # if any field lengths > 255, write csv file of _IDs and long fields
    if len(longFields) > 0 and isOpen:
        csvFile = open(outputDir+'/'+outName[0:-4]+'.txt','w')
        csvFile.write(fc+'_ID')
        for field in longFields:
                csvFile.write(','+field)
        csvFile.write('\n')
        rows = gp.searchCursor(fc)
        row = rows.next()
        while row:
            rowString = row.GetValue(fc+'_ID')
            for field in longFields:
                if row.GetValue(field) <> None:
                        rowString = rowString+'|'+row.GetValue(field)
                else:
                        rowString = rowString+'|'
            try:
                csvFile.write(rowString+'\n')	
            except:
                rStr = ''
                for i in range(len(rowString)):
                    if ord(rowString[i]) < 128:
                        rStr = rStr + rowString[i]
                    else: 
                        rStr = rStr + '<chr('+str(ord(rowString[i]))+')>'
                csvFile.write(rStr+'\n')
            row = rows.next()
        del row, rows
        csvFile.close()

def makeOutputDir(gdb,outWS,isOpen):
    outputDir = outWS+'/'+os.path.basename(gdb)[0:-4]
    if isOpen:
        outputDir = outputDir+'-open'
    else:
        outputDir = outputDir+'-simple'
    addMsgAndPrint(gp,'  Making '+outputDir+'/...')
    if os.path.exists(outputDir):
        if os.path.exists(outputDir+'/info'):
            for fl in glob.glob(outputDir+'/info/*'):
                os.remove(fl)
            os.rmdir(outputDir+'/info')
        for fl in glob.glob(outputDir+'/*'):
            os.remove(fl)
        os.rmdir(outputDir)
    os.mkdir(outputDir)
    logfile = open(outputDir+'/logfile.txt','w')
    logfile.write('file written by ncgmp_TranslateToShp.py, '+versionString+'\n\n')
    return outputDir, logfile

def dummyVal(pTerm,pVal):
    if pVal == None:
        if pTerm in equivalentFraction:
            return equivalentFraction[pTerm]
        else:
            return 0.0
    else:
        return pVal

def description(unitDesc):
    unitDesc.sort()
    unitDesc.reverse()
    desc = ''
    for uD in unitDesc:
        if uD[3] == '': desc = desc+str(uD[4])+':'
        else:  desc = desc+uD[3]+':'
        desc = desc+uD[2]+'; '
    return desc[:-2]

def makeStdLithDict(gp):
    addMsgAndPrint(gp,'  Making StdLith dictionary...')
    stdLithDict = {}
    rows = gp.searchcursor('StandardLithology',"","","","MapUnit")
    row = rows.next()
    unit = row.GetValue('MapUnit')
    unitDesc = []
    pTerm = row.GetValue('ProportionTerm'); pVal = row.GetValue('ProportionValue')
    val = dummyVal(pTerm,pVal)
    unitDesc.append([val,row.GetValue('PartType'),row.GetValue('Lithology'),pTerm,pVal])
    while row:
        #print row.GetValue('MapUnit')+'  '+row.GetValue('Lithology')
        newUnit = row.GetValue('MapUnit')
        if newUnit <> unit:
            stdLithDict[unit] = description(unitDesc)
            unitDesc = []
            unit = newUnit
        pTerm = row.GetValue('ProportionTerm'); pVal = row.GetValue('ProportionValue')
        val = dummyVal(pTerm,pVal)
        unitDesc.append([val,row.GetValue('PartType'),row.GetValue('Lithology'),pTerm,pVal])
        row = rows.next()
    del row, rows
    stdLithDict[unit] = description(unitDesc)
    return stdLithDict

def mapPolys(gp,stdLithDict,outputDir,logfile):
    addMsgAndPrint(gp,'  Translating GeologicMap/MapUnitPolys...')
    try:
        gp.MakeTableView_management('DescriptionOfMapUnits','DMU')
        gp.AddField_management('DMU',"StdLith","TEXT",'','','255')
        rows = gp.UpdateCursor('DMU'  )
        row = rows.Next()
        while row:
            if row.MapUnit in stdLithDict:
                row.StdLith = stdLithDict[row.MapUnit]
                rows.UpdateRow(row)
            row = rows.Next()
        del row, rows
        gp.MakeFeatureLayer_management("GeologicMap/MapUnitPolys","MUP")
        gp.AddJoin_management('MUP','MapUnit','DMU','MapUnit')
        gp.AddJoin_management('MUP','DataSourceID','DataSources','DataSources_ID')
        gp.CopyFeatures_management('MUP','MUP2')
        DM = 'DescriptionOfMapUnits_'
        DS = 'DataSources_'
        for field in ('DataSourceID','MapUnit_1','Label_1', DM+'OBJECTID',  
                           DM+DM+'ID', DM+'DefinitionSourceID',DS+'DefinitionSourceID',
                           DS+'OBJECTID',DS+DS+'ID'):
            gp.DeleteField_management('MUP2',field)
        dumpTable(gp,'MUP2','MapUnitPolys.shp',True,outputDir,logfile,False)
    except:
        addMsgAndPrint(gp,gp.GetMessages())
        addMsgAndPrint(gp,'  Failed to translate MapUnitPolys')

def linesAndPoints(gp,fc,outputDir,logfile):
    addMsgAndPrint(gp,'  Translating '+fc+'...')
    cp = fc.find('/')
    fcShp = fc[cp+1:]+'.shp'
    LIN2 = fc[cp+1:]+'2'
    try:
        gp.MakeFeatureLayer_management(fc,'LIN')
        gp.AddField_management('LIN','Definition','TEXT','#','#','255')
        gp.AddJoin_management('LIN','Type','Glossary','Term')
        gp.CalculateField_management('LIN','Definition','[Glossary.Definition]','VB')
        gp.RemoveJoin_management('LIN','Glossary')
        # commands below are 9.3+ specific
        sourceFields = gp.ListFields(fc,'*SourceID')
        for sField in sourceFields:
            nFieldName = sField.Name[:-2]
            gp.AddField_management('LIN',nFieldName,'TEXT','#','#','255')
            gp.AddJoin_management('LIN',sField.Name,'DataSources','DataSources_ID')
            gp.CalculateField_management('LIN',nFieldName,'[DataSources.Source]','VB')
            gp.RemoveJoin_management('LIN','DataSources')
            gp.DeleteField_management('LIN',sField.Name)
        gp.CopyFeatures_management('LIN',LIN2)
        gp.DeleteFeatures_management('LIN')
        dumpTable(gp,LIN2,fcShp,True,outputDir,logfile,False)
    except:
        addMsgAndPrint(gp,gp.GetMessages())
        addMsgAndPrint(gp,'  Failed to translate '+fc)


def main(gp,gdbCopy,outWS,oldgdb):
    #
    # Simple version
    #
    isOpen = False
    addMsgAndPrint(gp,'')
    outputDir, logfile = makeOutputDir(oldgdb,outWS,isOpen)
    # point feature classes
    gp.workspace = gdbCopy
    stdLithDict = makeStdLithDict(gp)
    mapPolys(gp,stdLithDict,outputDir,logfile)
    for fc in ('GeologicMap/ContactsAndFaults', 
                 'GeologicMap/ConcealedContactsAndFaults', 
                 'GeologicMap/OtherLines'):
            linesAndPoints(gp,fc,outputDir,logfile)
    # second line below requires 9.3+
    gp.workspace = 'GeologicMap'
    fcs = gp.ListFeatureClasses('','POINT')
    gp.workspace = gdbCopy
    for fc in fcs:
            linesAndPoints(gp,fc,outputDir,logfile)	
    logfile.close()
    #
    # Open version
    #
    isOpen = True
    addMsgAndPrint(gp,'')
    outputDir, logfile = makeOutputDir(oldgdb,outWS,isOpen)
    gp.workspace = gdbCopy
    # list featuredatasets
    fds = gp.ListDataSets()
    # for each featuredataset
    for fd in fds:
        addMsgAndPrint(gp, '  Processing feature data set '+fd+'...')
        logfile.write('Feature data set '+fd+' \n')
        spatialRef = gp.Describe(fd).SpatialReference
        logfile.write('  spatial reference framework\n')
        logfile.write('    name = '+spatialRef.Name+'\n')
        logfile.write('    spheroid = '+spatialRef.SpheroidName+'\n')
        logfile.write('    projection = '+spatialRef.ProjectionName+'\n')
        logfile.write('    units = '+spatialRef.LinearUnitName+'\n')
        # generate featuredataset prefix
        pfx = ''
        for i in range(0,len(fd)-1):
            if fd[i] == fd[i].upper():
                pfx = pfx + fd[i]
        # for each featureclass in dataset
        gp.workspace = gdbCopy
        gp.workspace = fd
        for fc in gp.ListFeatureClasses():
            outName = pfx+'_'+fc+'.shp'
            dumpTable(gp,fc,outName,True,outputDir,logfile,isOpen)
        logfile.write('\n')
    # list tables
    gp.workspace = gdbCopy
    for tbl in gp.listTables():
        outName = tbl+'.dbf'
        dumpTable(gp,tbl,outName,False,outputDir,logfile,isOpen)
    logfile.close()

### START HERE ###
print '  '+versionString
print '  checking input'
if len(sys.argv) <> 3 or not os.path.exists(sys.argv[1]) or not os.path.exists(sys.argv[2]):
    usage()
else:
    try:
        print '  loading arcprocessor'
        if Is93:
            gp = arcgisscripting.create(9.3)
        else:
            gp = arcgisscripting.create()
    except:
        addMsgAndPrint(gp,'  Unable to create geoprocessing object. Arc license not present??')
    else:
        gdb = os.path.abspath(sys.argv[1])
        ows = os.path.abspath(sys.argv[2])
        gp.overwriteoutput = True
        ## fix the new workspace name so it is guaranteed to be novel, no overwrite
        newgdb = ows+'/xx'+os.path.basename(gdb)
        addMsgAndPrint(gp,'  Copying '+os.path.basename(gdb)+' to temporary geodatabase...')
        gp.Copy_management(gdb,newgdb)
        main(gp,newgdb,ows,gdb)
        addMsgAndPrint(gp,'\n  Deleting temporary geodatabase...')
        addMsgAndPrint(gp,' NOTE THAT FOLLOWING ERROR MESSAGE IS EXPECTED')
        gp.Workspace = gdb
        time.sleep(5)
        gp.Delete_management(newgdb)

